--
-- Mutilate --
-- 

-- Our base array
Mutilate = {}

-- Mutilate variables NOT SAVED
Mutilate.versionNumber = 1.29
Mutilate.currentTarget = ""
Mutilate.lastTarget = ""
Mutilate.lastUpdateTime = 0


-- Return a localised spell name, including rank if any.
function FM_GetSpellNameById(spellId)
	if (spellId == nil) then
		return nil
	end
	local spellName, rank, _, _, _, _, _, _, _ = GetSpellInfo(spellId)
	if rank==nil then
		return spellName
	elseif string.len(rank) > 1 then
		return spellName.."("..rank..")"
	end
	return spellName
end

Mutilate.L = {}

Mutilate.E = {
    ["Vendetta"]            = 0,
	["Garrote"]             = 45,
	["Slice and Dice"]      = 25,
	["Mutilate"]            = 55,
	["Envenom"]             = 35,
	["Rupture"]             = 25,
	["Tricks of the Trade"] = 0,
	["Vanish"]              = 0,
	["Dispatch"]            = 30,
	["Blindside"]           = 0,
	["Recuperate"]          = 30,
	["Shiv"]          		= 20
}

Mutilate.textureList = {
  ["last"] = nil,
  ["current"] = nil,
  ["next"] = nil,
  ["misc"] = nil,
  ["int"] = nil,
  ["vendetta"] = nil,
}
	
local language = GetLocale();

Mutilate.L["Vendetta"]            = FM_GetSpellNameById(79140)
Mutilate.L["Garrote"]             = FM_GetSpellNameById(703)
Mutilate.L["Slice and Dice"]      = FM_GetSpellNameById(5171)
Mutilate.L["Mutilate"]            = FM_GetSpellNameById(1329)
Mutilate.L["Envenom"]             = FM_GetSpellNameById(32645)
Mutilate.L["Rupture"]             = FM_GetSpellNameById(1943)
Mutilate.L["Cold Blood"]          = FM_GetSpellNameById(14177)
Mutilate.L["Tricks of the Trade"] = FM_GetSpellNameById(57934)
Mutilate.L["Vanish"]              = FM_GetSpellNameById(1856)
Mutilate.L["Dispatch"]            = FM_GetSpellNameById(111240)
Mutilate.L["Blindside"]           = FM_GetSpellNameById(121152)
Mutilate.L["Kick"]           	  = FM_GetSpellNameById(1766)
Mutilate.L["Recuperate"]          = FM_GetSpellNameById(73651)
Mutilate.L["Improved Slice and Dice"] = FM_GetSpellNameById(157513)
Mutilate.L["Shiv"] 				  = FM_GetSpellNameById(5938)
Mutilate.L["Cloak of Shadows"] 	  = FM_GetSpellNameById(31224)

-- Balance druid
Mutilate.L["Lunar Empowerment"] 	  = FM_GetSpellNameById(164547)
Mutilate.L["Solar Empowerment"] 	  = FM_GetSpellNameById(164545)

Mutilate.lastUpdateTime = 0
Mutilate.playerName = UnitName("player")
Mutilate.spellHaste = GetCombatRatingBonus(20)


-- Our sneaky frame to watch for events ... checks Mutilate.events[] for the function.  Passes all args.
Mutilate.eventFrame = CreateFrame("Frame")
Mutilate.eventFrame:SetScript("OnEvent", function(this, event, ...)
  Mutilate.events[event](...)
end)

Mutilate.eventFrame:RegisterEvent("ADDON_LOADED")
Mutilate.eventFrame:RegisterEvent("PLAYER_LOGIN")
Mutilate.eventFrame:RegisterEvent("PLAYER_ALIVE")
Mutilate.eventFrame:RegisterEvent("PLAYER_TARGET_CHANGED")
Mutilate.eventFrame:RegisterEvent("PLAYER_TALENT_UPDATE")
--Mutilate.eventFrame:RegisterEvent("SPELLS_CHANGED")


-- Define our Event Handlers here
Mutilate.events = {}

function Mutilate:CheckSpec()
	local currentSpec = GetSpecialization()
	local currentSpecID = currentSpec and select(1, GetSpecializationInfo(currentSpec)) or 0
	if currentSpecID == 259 then -- 259 is Assasination spec
		Mutilate.spec = "Assasination"
		Mutilate.displayFrame:Show() 
	elseif currentSpecID == 102 then
		Mutilate.spec = "Balance"
		Mutilate.displayFrame:Show() 
	else
		Mutilate.displayFrame:Hide() 
	end 	
end

function Mutilate.events.PLAYER_TALENT_UPDATE()
	Mutilate:CheckSpec()

	-- Note if the player has Shadow Focus talent.
	local talentId, name, texture, selected, available = GetTalentInfo(1, 3, 1)
	Mutilate.hasShadowFocusTalent = selected
	-- and Anticipation talent.
	local talentId, name, texture, selected, available = GetTalentInfo(6, 3, 1)
	Mutilate.hasAnticipation = selected
	local inactive_regen_per_sec, active_regen_per_sec = GetPowerRegen()
	Mutilate.combatEnergyRegenPerSec = active_regen_per_sec
	Mutilate.hasImprovedSliceAndDice = false
	local skillType, spellId = GetSpellBookItemInfo("Improved Slice and Dice")
	if skillType ~= nil then
		Mutilate.hasImprovedSliceAndDice = true
	end
end

function Mutilate.events.PLAYER_ALIVE()

	Mutilate.eventFrame:UnregisterEvent("PLAYER_ALIVE")
end

function Mutilate.events.PLAYER_LOGIN()

	Mutilate.playerName = UnitName("player");

end

function Mutilate.events.ADDON_LOADED(addon)

	--for k, v in pairs(Mutilate.L) do print(v, k) end

	if addon ~= "Mutilate" then 
		return 
	end

	local _,playerClass = UnitClass("player")
	if playerClass ~= "ROGUE" then  --and playerClass ~= "DRUID" then
		Mutilate.eventFrame:UnregisterEvent("PLAYER_ALIVE")
		Mutilate.eventFrame:UnregisterEvent("ADDON_LOADED")
		Mutilate.eventFrame:UnregisterEvent("PLAYER_LOGIN")
		Mutilate.eventFrame:UnregisterEvent("PLAYER_TARGET_CHANGED")
		Mutilate.eventFrame:UnregisterEvent("PLAYER_TALENT_UPDATE")
		return 
	end
	
	-- Default saved variables
	if not Mutilatedb then
		Mutilatedb = {} -- fresh start
	end
	if not Mutilatedb.scale then Mutilatedb.scale = 1 end
	if Mutilatedb.locked == nil then Mutilatedb.locked = false end
	if not Mutilatedb.x then Mutilatedb.x = 100 end
	if not Mutilatedb.y then Mutilatedb.y = 100 end
	if not Mutilatedb.SuggestVanish then Mutilatedb.SuggestVanish = false end
	if not Mutilatedb.SuggestVendetta then Mutilatedb.SuggestVendetta = false end
	if not Mutilatedb.SliceAndDiceTimer then Mutilatedb.SliceAndDiceTimer = 1.0 end
	if not Mutilatedb.RuptureTimer then Mutilatedb.RuptureTimer = 2.0 end
	if not Mutilatedb.RuptureClipTimer then Mutilatedb.RuptureClipTimer = 0.0 end
	if not Mutilatedb.PoolEnergyThreshold then Mutilatedb.PoolEnergyThreshold = 90 end
	if not Mutilatedb.RuptureCp then Mutilatedb.RuptureCp = 1 end
	if not Mutilatedb.EnvenomCp then Mutilatedb.EnvenomCp = 4 end
	if not Mutilatedb.PlayerHealthThreshold then Mutilatedb.PlayerHealthThreshold = 50 end
	if not Mutilatedb.EnvenomTimer then Mutilatedb.EnvenomTimer = 0 end
	if not Mutilatedb.BackstabEnvenomCp then Mutilatedb.BackstabEnvenomCp = 5 end
	if not Mutilatedb.SuggestMutAsStealthOpener then Mutilatedb.SuggestMutAsStealthOpener = true end
	if Mutilatedb.SuggestCloakOfShadows == nil then Mutilatedb.SuggestCloakOfShadows = true end
	if not Mutilatedb.MinDebuffDurationForCofs then Mutilatedb.MinDebuffDurationForCofs = 10 end
	if Mutilatedb.VoicePrompts == nil then Mutilatedb.VoicePrompts = true end


	-- Create GUI
	Mutilate:CreateGUI()
	Mutilate.displayFrame:SetScale(Mutilatedb.scale)


	-- Create Options Frame
	Mutilate:CreateOptionFrame()
	if Mutilatedb.locked then
		Mutilate.displayFrame:SetScript("OnMouseDown", nil)
		Mutilate.displayFrame:SetScript("OnMouseUp", nil)
		Mutilate.displayFrame:SetScript("OnDragStop", nil)
		Mutilate.displayFrame:SetBackdropColor(0, 0, 0, 0)
		Mutilate.displayFrame:EnableMouse(false)
	else
		Mutilate.displayFrame:SetScript("OnMouseDown", function(self) self:StartMoving() end)
		Mutilate.displayFrame:SetScript("OnMouseUp", function(self) self:StopMovingOrSizing() end)
		Mutilate.displayFrame:SetScript("OnDragStop", function(self) self:StopMovingOrSizing() end)
		Mutilate.displayFrame:SetBackdropColor(0, 0, 0, .4)
		Mutilate.displayFrame:EnableMouse(true)
	end

	-- Register for Slash Commands
	SlashCmdList["Mutilate"] = Mutilate.Options
	SLASH_Mutilate1 = "/Mutilate"
	SLASH_Mutilate2 = "/mut"

	-- Register for Function Events
	Mutilate.eventFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	Mutilate.eventFrame:RegisterEvent("PLAYER_TARGET_CHANGED")

	--Mutilate.events.PLAYER_TALENT_UPDATE()

	Mutilate:CheckSpec()

	Mutilate.puid = UnitGUID("player")

	-- This is probably normally caled by framework, but just be be sure. Applies spec and talent based settings.
	Mutilate.events.PLAYER_TALENT_UPDATE()
	
end

-- Show or hide the spell textures.
function Mutilate:ShowSpellsIf(yes)
	if not yes then
		Mutilate.displayFrame_last:Hide()
		Mutilate.displayFrame_current:Hide()
		Mutilate.displayFrame_next:Hide()
		Mutilate.displayFrame_misc:Hide()
		Mutilate.displayFrame_int:Hide()
		Mutilate.displayFrame_vendetta:Hide()
		Mutilate.displayFrame_kick:Hide()
		Mutilate.displayFrame_cofs:Hide()
		Mutilate.kickNameFontString:SetText("")
		Mutilate.cofsNameFontString:SetText("")
		Mutilate.saidSpell = nil
		Mutilate.saidKick = false
	else
		Mutilate.displayFrame_last:Show()
		Mutilate.displayFrame_current:Show()
		Mutilate.displayFrame_next:Show()
		Mutilate.displayFrame_misc:Show()
		Mutilate.displayFrame_int:Show()
		Mutilate.displayFrame_vendetta:Show()
		Mutilate.displayFrame_kick:Show()
		Mutilate.displayFrame_cofs:Show()
		Mutilate:DecideSpells()
	end
end

function Mutilate.events.COMBAT_LOG_EVENT_UNFILTERED(timestamp, event, hideCaster, srcGUID, srcName, srcFlags, srcRaidFlags, dstGUID, dstName, dstFlags, destRaidFlags, ...)

	Mutilate:ShowSpellsIf(UnitName("target") ~= nil and not UnitIsFriend("player", "target") and UnitHealth("target") > 0)

end

function Mutilate.events.PLAYER_TARGET_CHANGED(...)
	
	-- target changed, set last target, update current target, will be nil if no target
	Mutilate.lastTarget = Mutilate.currentTarget
	Mutilate.currentTarget = UnitGUID("target")
	Mutilate:ShowSpellsIf(UnitName("target") ~= nil and not UnitIsFriend("player", "target") and UnitHealth("target") > 0)

end

-- End Event Handlers

function Mutilate:CreateGUI()

  local displayFrame = CreateFrame("Frame","MutilateDisplayFrame",UIParent)
  displayFrame:SetFrameStrata("BACKGROUND")
  displayFrame:SetWidth(350)
  displayFrame:SetHeight(107)
  displayFrame:SetBackdrop({
          bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", tile = true, tileSize = 32,
  			})
  displayFrame:SetBackdropColor(0, 0, 0, .4)
  displayFrame:EnableMouse(true)
  displayFrame:SetMovable(true)
  displayFrame:SetClampedToScreen(true)
  displayFrame:SetScript("OnMouseDown", function(self) self:StartMoving() end)
  displayFrame:SetScript("OnMouseUp", function(self) self:StopMovingOrSizing() end)
  displayFrame:SetScript("OnDragStop", function(self) self:StopMovingOrSizing() end)

  displayFrame:SetPoint("CENTER",-200,-200)

  local displayFrame_last = CreateFrame("Frame","$parent_last", MutilateDisplayFrame)
  local displayFrame_current = CreateFrame("Frame","$parent_current", MutilateDisplayFrame)
  local displayFrame_next = CreateFrame("Frame","$parent_next", MutilateDisplayFrame)
  local displayFrame_misc = CreateFrame("Frame","$parent_misc", MutilateDisplayFrame)
  local displayFrame_int = CreateFrame("Frame","$parent_int", MutilateDisplayFrame)
  local displayFrame_vendetta = CreateFrame("Frame","$parent_vendetta", MutilateDisplayFrame)
  local displayFrame_kick = CreateFrame("Frame","$parent_vendetta", MutilateDisplayFrame)
  local displayFrame_cofs = CreateFrame("Frame","$parent_vendetta", MutilateDisplayFrame)
  local kickNameFontString = displayFrame:CreateFontString("kickName", "OVERLAY", "GameFontNormal")
  kickNameFontString:SetText("")
  kickNameFontString:SetPoint("TOPLEFT", 0, -95)
  local displayFrame_cofs = CreateFrame("Frame","$parent_vendetta", MutilateDisplayFrame)
  local cofsNameFontString = displayFrame:CreateFontString("cofsName", "OVERLAY", "GameFontNormal")
  cofsNameFontString:SetJustifyH("RIGHT")
  cofsNameFontString:SetText("")
  cofsNameFontString:SetPoint("TOPRIGHT", 0, -45)

  displayFrame_last:SetWidth(45)
  displayFrame_current:SetWidth(70)
  displayFrame_next:SetWidth(45)
  displayFrame_misc:SetWidth(45)
  displayFrame_int:SetWidth(45)
  displayFrame_vendetta:SetWidth(45)
  displayFrame_kick:SetWidth(45)
  displayFrame_cofs:SetWidth(45)

  displayFrame_last:SetHeight(45)
  displayFrame_current:SetHeight(70)
  displayFrame_next:SetHeight(45)
  displayFrame_misc:SetHeight(45)
  displayFrame_int:SetHeight(45)
  displayFrame_vendetta:SetHeight(45)
  displayFrame_kick:SetHeight(45)
  displayFrame_cofs:SetHeight(45)

  displayFrame_current:SetPoint("TOPLEFT", 140, -10)
  displayFrame_next:SetPoint("TOPLEFT", 250, 0) -- was -50
  
  displayFrame_last:SetPoint("TOPLEFT", 0, -50)

  displayFrame_misc:SetPoint("TOPLEFT", 0, 0)
  displayFrame_int:SetPoint("TOPLEFT", 50, 0)
  displayFrame_vendetta:SetPoint("TOPLEFT", 50,-50) 
  displayFrame_kick:SetPoint("TOPLEFT", 0,-50) 
  displayFrame_cofs:SetPoint("TOPLEFT", 300, 0) 


  local t = displayFrame_last:CreateTexture(nil,"BACKGROUND")
  t:SetTexture(nil)
  t:SetAllPoints(displayFrame_last)
  t:SetAlpha(.8)
  displayFrame_last.texture = t
  Mutilate.textureList["last"] = t

  t = displayFrame_current:CreateTexture(nil,"BACKGROUND")
  t:SetTexture(nil)
  t:ClearAllPoints()
  t:SetAllPoints(displayFrame_current)
  displayFrame_current.texture = t
  Mutilate.textureList["current"] = t

  t = displayFrame_misc:CreateTexture(nil,"BACKGROUND")
  t:SetTexture(nil)
  t:SetAllPoints(displayFrame_misc)
  t:SetAlpha(.8)
  displayFrame_misc.texture = t
  Mutilate.textureList["misc"] = t

  t = displayFrame_int:CreateTexture(nil,"BACKGROUND")
  t:SetTexture(nil)
  t:SetAllPoints(displayFrame_int)
  t:SetAlpha(.8)
  displayFrame_int.texture = t
  Mutilate.textureList["int"] = t

  t = displayFrame_next:CreateTexture(nil,"BACKGROUND")
  t:SetTexture(nil)
  t:SetAllPoints(displayFrame_next)
  t:SetAlpha(.8)
  displayFrame_next.texture = t
  Mutilate.textureList["next"] = t
  
  t = displayFrame_vendetta:CreateTexture(nil,"BACKGROUND")
  t:SetTexture(nil)
  t:SetAllPoints(displayFrame_vendetta)
  t:SetAlpha(.8)
  displayFrame_vendetta.texture = t
  Mutilate.textureList["vendetta"] = t

  t = displayFrame_kick:CreateTexture(nil,"BACKGROUND")
  t:SetTexture(nil)
  t:SetAllPoints(displayFrame_kick)
  t:SetAlpha(.8)
  displayFrame_kick.texture = t
  Mutilate.textureList["kick"] = t

  t = displayFrame_cofs:CreateTexture(nil,"BACKGROUND")
  t:SetTexture(nil)
  t:SetAllPoints(displayFrame_cofs)
  t:SetAlpha(.8)
  displayFrame_cofs.texture = t
  Mutilate.textureList["cofs"] = t

  displayFrame:SetScript("OnUpdate", function(this, elapsed)
    Mutilate:OnUpdate(elapsed)
  end)

  local cooldownFrame = CreateFrame("Cooldown", "$parent_cooldown", MutilateDisplayFrame)
  cooldownFrame:SetFrameLevel(3)
  cooldownFrame:SetPoint("CENTER", displayFrame_current, "CENTER", 0, 0)
  cooldownFrame:SetHeight(70)
  cooldownFrame:SetWidth(70)
  cooldownFrame:ClearAllPoints()
  
  Mutilate.displayFrame = displayFrame
  Mutilate.displayFrame_last = displayFrame_last
  Mutilate.displayFrame_current = displayFrame_current
  Mutilate.displayFrame_next = displayFrame_next
  Mutilate.displayFrame_misc =  displayFrame_misc
  Mutilate.displayFrame_int =  displayFrame_int
  Mutilate.displayFrame_vendetta = displayFrame_vendetta
  Mutilate.displayFrame_kick = displayFrame_kick
  Mutilate.displayFrame_cofs = displayFrame_cofs
  Mutilate.cooldownFrame = cooldownFrame
  Mutilate.kickNameFontString = kickNameFontString
  Mutilate.cofsNameFontString = cofsNameFontString

end

function Mutilate:OnUpdate(elapsed)
		Mutilate:DecideSpells()
end

function Mutilate:GetBalanceState(now)

	local globalCoolDownExpiration = now
	local start, duration, enabled = GetSpellCooldown(61304); -- The speical global cooldown spell ID.
	if duration ~= nil and duration ~= 0 then
		globalCoolDownExpiration = max(now, start + duration)
		if now - start < 0.3 then 
			return nil
		end
	end
	local s = {}
	s.mana = UnitPower("player", 0)
	s.eclipsePower = UnitPower("player", 8)
	s.isInGroup = IsInGroup()
	local mastery, coefficient = GetMasteryEffect()
	--print("power=", s.power, " GetEclipseDirection()=", GetEclipseDirection(), " master=", mastery, " coefficient=", coefficient)

	s.lunarEmpowermentExpiration = 0
	s.lunarEmpowermentCount = 0
	local name, rank, icon, count, debuffType, duration, expirationTime, isMine, isStealable = UnitBuff("player", Mutilate.L["Lunar Empowerment"])
	if name ~= nil then
		s.lunarEmpowermentExpiration = expirationTime
		s.lunarEmpowermentCount = count
	end

	s.solarEmpowermentExpiration = 0
	s.solarEmpowermentCount = 0
	local name, rank, icon, count, debuffType, duration, expirationTime, isMine, isStealable = UnitBuff("player", Mutilate.L["Solar Empowerment"])
	if name ~= nil then
		s.solarEmpowermentExpiration = expirationTime
		s.solarEmpowermentCount = count
	end

	--print(Mutilate:StateSummary(s))
	return nil

end

function Mutilate:GetBalanceSpellFromState(s)
	local spell = nil;
	local why = "no case";
	local oldNow = s.now

	if s.globalCoolDownExpiration > s.now then

		why = "waiting for global cooldown"
		s.now = s.globalCoolDownExpiration


	--Starsurg
	--if in lunar and no lunarEmpowerment charges and starsurge avail and 3.bit wrath cast times of remaing then starsurge 
	--if in solar and no solarEmpowerment charges and xxx avail and 2.bit starfire cast times of remaing then xxx 

		--same for solarEmpowe
	--elseif s.targetIsEnraged and s.energy > Mutilate.E["Shiv"] * eCostMult then

	end
end

-- Return a table with all the interesting facts about the current state. All times are absolute.
-- Ready times take into account global cooldown. Buffs and debuffs are active if their expiration
-- is in the future. Assumes target existis and is unfriendly.
-- Returns nil if we are in first few hundred milliseconds of the start of global cooldown, in
-- which state is often confused. In this situation, caller should assume a spell has been cast
-- (i.e. clear the recommendation) and hold predition as is.
function Mutilate:GetState(now)

	local globalCoolDownExpiration = now
	local start, duration, enabled = GetSpellCooldown(61304); -- The speical global cooldown spell ID.
	if duration ~= nil and duration ~= 0 then
		globalCoolDownExpiration = max(now, start + duration)
		if now - start < 0.3 then 
			return nil
		end
	end

	local s = {}
	s.now = now
	s.globalCoolDownExpiration = globalCoolDownExpiration

	s.energy = UnitPower("player")
	s.comboPoints = GetComboPoints("player")
	s.isInGroup = IsInGroup()

	-- Scan taget's buffs for an enrage effect that Shiv could remove.
	s.targetIsEnraged = false
	for i = 1, 40 do -- There are 1..40 aura slots.
		local name, _, _, _, debuffType, _, _, _, canStealOrPurge, _, _, _, _, _ = UnitAura("target", i)
		-- E.g. Hozen Rage icon 0 "" 12 530203.832 target true false 115006 false false false
		if name == nil then
			break -- Assuming here that there a no gaps. There is no GetNumAuras function I believe.
		end
		if debuffType == "" and canStealOrPurge then
			s.targetIsEnraged = true -- Empty string is special, means enrage.
		end
	end

	s.removeableName = ""
	s.longDurationDebuffs = false
	-- Scan player's debuffs for things Cloak of Shadows could remove.
	for i = 1, 40 do -- There are 1..40 aura slots.
		local name, _, _, count, debuffType, _, expirationTime, unitCaster, canStealOrPurge, _, _, _, _ = UnitDebuff("player", i)
		if name == nil or unitCaster == "player" then
			break 
		end
		if expirationTime > s.now then
			local duration = math.floor(expirationTime - s.now)
			if duration < 60 * 10 then -- Don't show times > 10m
				s.removeableName = s.removeableName..""..duration.."s "
			end
			if duration > Mutilatedb.MinDebuffDurationForCofs then
				s.longDurationDebuffs = true
			end
		else
			s.longDurationDebuffs = true
		end
		if count > 1 then
			s.removeableName = s.removeableName..count.."*"
		end
		s.removeableName = s.removeableName..name.."\n"
	end

	s.interruptibleName = nil
	local name, _, displayName, _, _, endTime, _, notInterruptible = UnitChannelInfo("target")
	if name ~= nil and not notInterruptible then
		s.interruptibleName = displayName
	end
	local name, _, displayName, _, _, endTime, _, _, notInterruptible = UnitCastingInfo("target")
	if name ~= nil and not notInterruptible then
		s.interruptibleName = displayName
	end

	s.garroteExpiration = 0
	local name, rank, icon, count, debuffType, duration, expirationTime, isMine, isStealable = UnitDebuff("target", Mutilate.L["Garrote"])
	if name ~= nil and isMine == "player" then
		s.garroteExpiration = expirationTime
	end	

	s.sliceNDiceExpiration = s.now;
	if (Mutilate.hasImprovedSliceAndDice) then
		s.sliceNDiceExpiration = s.sliceNDiceExpiration + 1000 -- Model Improved SnD as long running SnD
	else
		local name, rank, icon, count, debuffType, duration, expirationTime, isMine, isStealable = UnitBuff("player", Mutilate.L["Slice and Dice"])
		if name ~= nil then
			s.sliceNDiceExpiration = expirationTime
		end
	end

	s.ruptureExpiration = 0
	local name, rank, icon, count, debuffType, duration, expirationTime, isMine, isStealable = UnitDebuff("target", Mutilate.L["Rupture"])
	if name ~= nil and isMine == "player" then
		s.ruptureExpiration = expirationTime
	end	

	s.vendettaReadyTime = globalCoolDownExpiration
	local start, duration, enabled = GetSpellCooldown(Mutilate.L["Vendetta"]);
	if duration ~= nil then 
		s.vendettaReadyTime = max(start + duration, globalCoolDownExpiration)
	end

	s.tricksOfTheTradeReadyTime = globalCoolDownExpiration
	local start, duration, enabled = GetSpellCooldown(Mutilate.L["Tricks of the Trade"]);
	if duration ~= nil then 
		s.tricksOfTheTradeReadyTime = max(start + duration, globalCoolDownExpiration)
	end

	s.vanishReadyTime = globalCoolDownExpiration
	local start, duration, enabled = GetSpellCooldown(Mutilate.L["Vanish"]);
	if duration ~= nil then
		s.vanishReadyTime = max(start + duration, globalCoolDownExpiration)
	end	

	s.kickReadyTime = globalCoolDownExpiration
	local start, duration, enabled = GetSpellCooldown(Mutilate.L["Kick"]);
	if duration ~= nil then
		s.kickReadyTime = max(start + duration, globalCoolDownExpiration)
	end	

	s.cofsReadyTime = globalCoolDownExpiration
	local start, duration, enabled = GetSpellCooldown(Mutilate.L["Cloak of Shadows"]);
	if duration ~= nil then
		s.cofsReadyTime = max(start + duration, globalCoolDownExpiration)
	end	

	s.envenomExpiration = 0
	local name, rank, icon, count, debuffType, duration, expirationTime, isMine, isStealable = UnitBuff("player", Mutilate.L["Envenom"])
	if duration ~= nil then
		s.envenomExpiration = expirationTime 
	end

	s.blindSideExpiration = 0
	local name, rank, icon, count, debuffType, duration, expirationTime, isMine, isStealable = UnitBuff("player", Mutilate.L["Blindside"])
	if name ~= nil then
		s.blindSideExpiration = expirationTime
	end

	s.recuperateExpiration = 0
	local name, rank, icon, count, debuffType, duration, expirationTime, isMine, isStealable = UnitBuff("player", Mutilate.L["Recuperate"])
	if name ~= nil then
		s.recuperateExpiration = expirationTime
	end

	s.playerIsStealthed = UnitAura("player", "Stealth") ~= nil

	s.targetHealthFraction = UnitHealth("target") / UnitHealthMax("target")

	s.playerHealthFraction = UnitHealth("player") / UnitHealthMax("player")

	--for k, v in pairs(s) do print(v, k) end
	return s
end

function Mutilate:StateSummary(s)
	local summary = ""
	for k, v in pairs(s) do
		summary = summary .. " " .. k .. "=" .. tostring(v)
	end
	return summary
end

-- Determine a recommened spell, if any, from the given state. In the process, advance the state to
-- some future time where something might be different. Always advances the time of the state by at
-- least a little.
-- Returned spell is not localised.
function Mutilate:GetSpellFromState(s, debugStr)
	-- http://www.icy-veins.com/wow/assassination-rogue-pve-dps-rotation-cooldowns-abilities
	-- As of 2014-11-01

	local spell = nil;
	local why = "no case";
	local oldNow = s.now
	local fullCombos = 4
	if (Mutilate.hasAnticipation) then
		fullCombos = 5
	end

	local eCostMult = 1
	if s.playerIsStealthed and Mutilate.hasShadowFocusTalent then
		eCostMult = 0.25
	end

	if s.globalCoolDownExpiration > s.now then

		why = "waiting for global cooldown"
		s.now = s.globalCoolDownExpiration 

	elseif s.targetIsEnraged and s.energy > Mutilate.E["Shiv"] * eCostMult then

		spell ="Shiv"
		why = "should dispell the enrage"
		s.energy = s.energy - Mutilate.E["Shiv"] * eCostMult
		s.targetIsEnraged = false
		s.playerIsStealthed = false

	elseif s.energy >= Mutilate.E["Recuperate"] * eCostMult
		and s.recuperateExpiration < s.now
		and s.playerHealthFraction < Mutilatedb.PlayerHealthThreshold / 100
		and s.comboPoints >= 1
		then

		spell = "Recuperate"
		s.recuperateExpiration = s.now + s.comboPoints * 6
		s.comboPoints = 0
		s.energy = s.energy - Mutilate.E["Recuperate"] * eCostMult
		why = "you're dying"

	elseif s.sliceNDiceExpiration <= s.now and s.comboPoints > 0 and s.energy >= Mutilate.E["Slice and Dice"] * eCostMult then

		spell = "Slice and Dice";
		why = "should always have it up, and you have combo points for it"
		s.sliceNDiceExpiration = s.now + s.comboPoints * 8 -- Approx
		s.comboPoints = 0
		s.energy = s.energy - Mutilate.E["Slice and Dice"] * eCostMult
		s.playerIsStealthed = false

	elseif s.energy >= Mutilate.E["Rupture"] * eCostMult
		and ((s.ruptureExpiration < s.now + 8 and s.comboPoints >= fullCombos)
			or (s.ruptureExpiration < s.now + 2 and s.comboPoints >= 1)) then

		-- "It is optimal to apply and refresh, with 8 seconds or less remaining, Rupture with 5 Combo Points.
		-- That said, if it is about to expire, refresh it with whatever amount of Combo Points you have."
		spell = "Rupture"
		why = "it needs refreshing"
		s.ruptureExpiration = s.now + s.comboPoints * 5 -- Approx
		s.comboPoints = 0
		s.energy = s.energy - Mutilate.E["Rupture"] * eCostMult
		s.playerIsStealthed = false

	elseif s.energy >= Mutilate.E["Envenom"] * eCostMult
		and s.energy >= UnitPowerMax("player") * 0.95
		and s.comboPoints > 0 then

		spell = "Envenom"
		why = "it is better than wasting energy"
		s.comboPoints = 0
		s.energy = s.energy - Mutilate.E["Envenom"] * eCostMult
		s.envenomExpiration = s.now + s.comboPoints * 2 -- Approx
		s.playerIsStealthed = false

	elseif s.energy >= Mutilate.E["Envenom"] * eCostMult and
		(s.comboPoints >= fullCombos or (s.comboPoints >= 1 and s.sliceNDiceExpiration < s.now + 3)) then

		-- "Use Envenom with 5 Combo Points. If Slice and Dice needs refreshing,
		-- then you can afford to cast Envenom with fewer Combo Points."
		spell = "Envenom"
		why = "it needs refreshing"
		s.comboPoints = 0
		s.energy = s.energy - Mutilate.E["Envenom"] * eCostMult
		s.envenomExpiration = s.now + s.comboPoints * 2 -- Approx
		s.playerIsStealthed = false

	elseif s.comboPoints < 5 then

		if s.energy >= Mutilate.E["Dispatch"] * eCostMult
			and (s.blindSideExpiration > s.now or s.targetHealthFraction < 0.35) then

			spell = "Dispatch"
			why = "you need to build combo point and low target health (or blindside proc)"
			s.blindSideExpiration = s.now
			s.comboPoints = s.comboPoints + 1
			s.energy = s.energy - Mutilate.E["Dispatch"] * eCostMult
			s.playerIsStealthed = false

		elseif s.energy >= Mutilate.E["Mutilate"] * eCostMult then

			spell = "Mutilate"
			why = "you need to build combo points"
			s.comboPoints = s.comboPoints + 2
			s.energy = s.energy - Mutilate.E["Mutilate"] * eCostMult
			s.playerIsStealthed = false

		end
	end

	local vanishOk = s.vanishReadyTime <= s.now and s.isInGroup

	local vendettaOk = s.vendettaReadyTime <= s.now

	local tricksOfTheTradeOk = s.tricksOfTheTradeReadyTime <= s.now and s.isInGroup

	local kickOk = s.kickReadyTime <= s.now and s.interruptibleName ~= nil

	local cofsOk = s.cofsReadyTime <= s.now and s.removeableName ~= nil

	-- Update state to next intersting time.
	if spell ~= nil then
		s.now = max(s.now, oldNow + 1) -- global cool down
		s.globalCoolDownExpiration = s.now
	end
	s.now = max(s.now, oldNow + 0.2) -- Always move forward in time at least a little
	s.energy = s.energy + Mutilate.combatEnergyRegenPerSec * (s.now - oldNow)

	return spell, vanishOk, vendettaOk, tricksOfTheTradeOk, kickOk, s.interruptibleName, s.longDurationDebuffs, s.removeableName, why
end

-- spell is a non-localised name.
function Mutilate:saySpell(spell)
	if Mutilatedb.VoicePrompts then
		-- Try to play localised version, is present.
		local willPlay = PlaySoundFile("Interface\\AddOns\\Mutilate\\"..Mutilate.L[spell]..".mp3", "Master")
		if not willPlay and Mutilate.L[spell] ~= spell then
			-- Failing that, play English (if it is different)
			PlaySoundFile("Interface\\AddOns\\Mutilate\\"..spell..".mp3", "Master")
		end
	end
end

function Mutilate:DecideSpells()

	local now = GetTime()
	if now - Mutilate.lastUpdateTime < 0.05 then
		return
	end
	Mutilate.lastUpdateTime = now
	
	local guid = UnitGUID("target")
	--local puid = UnitGUID("player")
	
	if  UnitName("target") == nil or UnitIsFriend("player", "target") or UnitHealth("target") == 0 then
		return -- ignore the dead and friendly
	end

	if guid == nil then
		Mutilate.textureList["last"]:SetTexture(nil)
		Mutilate.textureList["current"]:SetTexture(nil)
		Mutilate.textureList["next"]:SetTexture(nil)
		Mutilate.textureList["misc"]:SetTexture(nil)
		Mutilate.textureList["int"]:SetTexture(nil)
		Mutilate.textureList["vendetta"]:SetTexture(nil)
		Mutilate.textureList["kick"]:SetTexture(nil)
		Mutilate.textureList["cofs"]:SetTexture(nil)
		Mutilate.saidSpell = nil
		Mutilate.saidKick = false
		return
	end

	local s;
	--print("spec=", Mutilate.spec)
	--if Mutilate.spec == "Assasination" then
		s = Mutilate:GetState(now)
	--else
		--s = Mutilate:GetBalanceState(now)
	--end
	if s == nil then
		-- In period of confusion following starb of global cooldown.
		Mutilate.textureList["current"]:SetTexture(nil)
		return
	end

	local globalCoolDownExpiration = s.globalCoolDownExpiration
	local spell, vanishOk, vendettaOk, tricksOfTheTradeOk, kickOk, interruptibleName, cofsOk, removeableName, why = Mutilate:GetSpellFromState(s, "=")
	if spell == nil then
		Mutilate.textureList["current"]:SetTexture(nil)
		Mutilate.saidSpell = nil
	else
		Mutilate.textureList["current"]:SetTexture(GetSpellTexture(Mutilate.L[spell]))
		if Mutilate.saidSpell ~= spell then
			Mutilate:saySpell(spell)
			Mutilate.saidSpell = spell
		end
	end
	
	if tricksOfTheTradeOk then
		Mutilate.textureList["misc"]:SetTexture(GetSpellTexture(Mutilate.L["Tricks of the Trade"]))
	else
		Mutilate.textureList["misc"]:SetTexture(nil)
	end

	if vanishOk then
		Mutilate.textureList["int"]:SetTexture(GetSpellTexture(Mutilate.L["Vanish"]))
	else
		Mutilate.textureList["int"]:SetTexture(nil)
	end
	
	if vendettaOk then
		Mutilate.textureList["vendetta"]:SetTexture(GetSpellTexture(Mutilate.L["Vendetta"]))
	else
		Mutilate.textureList["vendetta"]:SetTexture(nil)
	end

	if kickOk then
		Mutilate.textureList["kick"]:SetTexture(GetSpellTexture(Mutilate.L["Kick"]))
		Mutilate.kickNameFontString:SetText("|cffFFffFF"..interruptibleName)
		if not Mutilate.saidKick then
			Mutilate:saySpell("Kick")
			Mutilate.saidKick = true
		end
	else
		Mutilate.textureList["kick"]:SetTexture(nil)
		Mutilate.kickNameFontString:SetText("")
		Mutilate.saidKick = false
	end

	if cofsOk then
		Mutilate.textureList["cofs"]:SetTexture(GetSpellTexture(Mutilate.L["Cloak of Shadows"]))
		Mutilate.cofsNameFontString:SetText("|cffFFffFF"..removeableName)
	else
		Mutilate.textureList["cofs"]:SetTexture(nil)
		Mutilate.cofsNameFontString:SetText("")
	end

	-- Try to predict next required spell using extrapolation of state.
	local i = 0
	local nextSpell = nil
	local whenNext = now
	while i < 20 do
		whenNext = s.now
		local nspell, vanishOk, vendettaOk, tricksOfTheTradeOk, kickOk, interruptibleName, cofsOk, removeableName, why = Mutilate:GetSpellFromState(s, "+")
		nextSpell = nspell
		if nextSpell ~= nil then
			break
		end
		i = i + 1
	end
	if nextSpell == nil then 
		Mutilate.textureList["next"]:SetTexture(nil)
	else
		Mutilate.textureList["next"]:SetTexture(GetSpellTexture(Mutilate.L[nextSpell]))
	end

end

function Mutilate:CreateOptionFrame()
	local panel = CreateFrame("FRAME", "MutilateOptions");
	panel.name = "Mutilate";

	local y = -10
	local lineStep = 30

	local fstring1 = panel:CreateFontString("MutilateOptions_string1", "OVERLAY", "GameFontNormal")
	fstring1:SetText("Lock panel position")
	fstring1:SetPoint("TOPLEFT", 10, y)

	local checkbox1 = CreateFrame("CheckButton", "$parent_cb1", panel, "OptionsCheckButtonTemplate")
	checkbox1:SetWidth(18)
	checkbox1:SetHeight(18)
	checkbox1:SetScript("OnClick", function() Mutilate:ToggleLocked() end)
	checkbox1:SetPoint("TOPRIGHT", -10, y)
	checkbox1:SetChecked(Mutilate:GetLocked())
	y = y - lineStep
	
	--
	-- GUI Scale
	--
	local fstring5 = panel:CreateFontString("MutilateOptions_string5", "OVERLAY", "GameFontNormal")
	fstring5:SetText("GUI Scale")
	fstring5:SetPoint("TOPLEFT", 10, y)

	local slider5 = CreateFrame("Slider", "$parent_sl", panel, "OptionsSliderTemplate")
	slider5:SetMinMaxValues(.5, 1.5)
	slider5:SetValue(Mutilatedb.scale)
	slider5:SetValueStep(1.0 / 16)
	slider5:SetObeyStepOnDrag(true)
	slider5:SetScript("OnValueChanged", function(self) Mutilate:SetScale(self:GetValue()); getglobal(self:GetName() .. "Text"):SetText(self:GetValue())  end)
	getglobal(slider5:GetName() .. "Low"):SetText("0.5")
	getglobal(slider5:GetName() .. "High"):SetText("1.5")
	getglobal(slider5:GetName() .. "Text"):SetText(Mutilatedb.scale)
	slider5:SetPoint("TOPRIGHT", -10, y)
	y = y - 2 * lineStep
	
	--
	-- Recuperate player health threshold
	--
	local fstring11 = panel:CreateFontString("MutilateOptions_string11", "OVERLAY", "GameFontNormal")
	fstring11:SetText("Suggest Recuperate when your health is below this %")
	fstring11:SetPoint("TOPLEFT", 10, y)

	local slider11 = CreateFrame("Slider", "$parent_s7", panel, "OptionsSliderTemplate")
	slider11:SetMinMaxValues(0, 100)
	slider11:SetValue(Mutilatedb.PlayerHealthThreshold)
	slider11:SetValueStep(1)
	slider11:SetObeyStepOnDrag(true)
	slider11:SetScript("OnValueChanged", function(self) Mutilate:SetPlayerHealthThreshold(self:GetValue()); getglobal(self:GetName() .. "Text"):SetText(self:GetValue())  end)
	getglobal(slider11:GetName() .. "Low"):SetText("0")
	getglobal(slider11:GetName() .. "High"):SetText("100")
	getglobal(slider11:GetName() .. "Text"):SetText(Mutilatedb.PlayerHealthThreshold)
	slider11:SetPoint("TOPRIGHT", -10, y)
	y = y - lineStep

	-- Cloak of Shadows checkbox
	local fs = panel:CreateFontString("MutilateOptions_cofsCheck", "OVERLAY", "GameFontNormal")
	fs:SetText("Show Cloak of Shadows icon when you have debuffs")
	fs:SetPoint("TOPLEFT", 10, y)

	local cb = CreateFrame("CheckButton", "$parent_cb1", panel, "OptionsCheckButtonTemplate")
	cb:SetWidth(18)
	cb:SetHeight(18)
	cb:SetScript("OnClick", function() Mutilatedb.SuggestCloakOfShadows = not Mutilatedb.SuggestCloakOfShadows end)
	cb:SetPoint("TOPRIGHT", -10, y)
	cb:SetChecked(Mutilatedb.SuggestCloakOfShadows)
	y = y - lineStep
	
	local fs = panel:CreateFontString("MutilateOptions_cofsSlider", "OVERLAY", "GameFontNormal")
	fs:SetText("Only show Cloak of Shadows icon when a debuff has more duration than this")
	fs:SetPoint("TOPLEFT", 10, y)

	local slider = CreateFrame("Slider", "$parent_sCofs", panel, "OptionsSliderTemplate")
	slider:SetMinMaxValues(0, 60)
	slider:SetValue(Mutilatedb.MinDebuffDurationForCofs)
	slider:SetValueStep(1)
	slider:SetObeyStepOnDrag(true)
	slider:SetScript("OnValueChanged", function(self) Mutilatedb.MinDebuffDurationForCofs = self:GetValue(); getglobal(self:GetName() .. "Text"):SetText(self:GetValue())  end)
	getglobal(slider:GetName() .. "Low"):SetText("0")
	getglobal(slider:GetName() .. "High"):SetText("60")
	getglobal(slider:GetName() .. "Text"):SetText(Mutilatedb.MinDebuffDurationForCofs)
	slider:SetPoint("TOPRIGHT", -10, y)
	y = y - lineStep

	local cb = CreateFrame("CheckButton", "$parent_cb1", panel, "OptionsCheckButtonTemplate")
	cb:SetWidth(18)
	cb:SetHeight(18)
	cb:SetScript("OnClick", function() Mutilatedb.VoicePrompts = not Mutilatedb.VoicePrompts end)
	cb:SetPoint("TOPRIGHT", -10, y)
	cb:SetChecked(Mutilatedb.VoicePrompts)
	
	local fs = panel:CreateFontString("MutilateOptions_cofsSlider", "OVERLAY", "GameFontNormal")
	fs:SetText("Play voice prompts for main recommendation and kick")
	fs:SetPoint("TOPLEFT", 10, y)

	y = y - lineStep
	local fstring = panel:CreateFontString(nil, "OVERLAY", "GameFontNormal")
	fstring:SetText("|cffA0A0A0Logic based on:")
	fstring:SetPoint("TOPLEFT", 10, y)
	y = y - lineStep / 2

	local fstring = panel:CreateFontString(nil, "OVERLAY", "GameFontNormal")
	fstring:SetText("|cffA0A0A0http://www.icy-veins.com/wow/assassination-rogue-pve-dps-rotation-cooldowns-abilities")
	fstring:SetPoint("TOPLEFT", 10, y)
	y = y - lineStep

	local fstring = panel:CreateFontString(nil, "OVERLAY", "GameFontNormal")
	fstring:SetText("|cffA0A0A0Most options have been removed in favor of automatic settings. If you want some back,")
	fstring:SetPoint("TOPLEFT", 10, y)
	y = y - lineStep / 2

	local fstring = panel:CreateFontString(nil, "OVERLAY", "GameFontNormal")
	fstring:SetText("|cffA0A0A0or have other requests, leave a note at: http://www.curse.com/addons/wow/mutilate")
	fstring:SetPoint("TOPLEFT", 10, y)
	y = y - lineStep

	InterfaceOptions_AddCategory(panel);
end

function Mutilate:GetLocked()
	return Mutilatedb.locked
end

function Mutilate:ToggleLocked()
	if Mutilatedb.locked then
		Mutilatedb.locked = false
		Mutilate.displayFrame:SetScript("OnMouseDown", function(self) self:StartMoving() end)
		Mutilate.displayFrame:SetScript("OnMouseUp", function(self) self:StopMovingOrSizing() end)
		Mutilate.displayFrame:SetScript("OnDragStop", function(self) self:StopMovingOrSizing() end)
		Mutilate.displayFrame:SetBackdropColor(0, 0, 0, .4)
		Mutilate.displayFrame:EnableMouse(true)
	else
		Mutilatedb.locked = true
		Mutilate.displayFrame:SetScript("OnMouseDown", nil)
		Mutilate.displayFrame:SetScript("OnMouseUp", nil)
		Mutilate.displayFrame:SetScript("OnDragStop", nil)
		Mutilate.displayFrame:SetBackdropColor(0, 0, 0, 0)
		Mutilate.displayFrame:EnableMouse(false)
	end
end


function Mutilate.Options(msg, editBox)

	if(msg == 'show') then
		Mutilate.displayFrame:Show() 
	elseif(msg == 'hide') then
		Mutilate.displayFrame:Hide() 
	elseif(msg == '') then
		InterfaceOptionsFrame_OpenToCategory(getglobal("MutilateOptions"))
	end
	
end
	
function Mutilate:SetScale(num)
	Mutilatedb.scale = num
	Mutilate.displayFrame:SetScale(Mutilatedb.scale)
	Mutilate.cooldownFrame:SetScale(Mutilatedb.scale)
end

function Mutilate:SetPlayerHealthThreshold(num)
	Mutilatedb.PlayerHealthThreshold = num
end


function Mutilate:shallowCopyOfTable(t)
  local t2 = {}
  for k, v in pairs(t) do
    t2[k] = v
  end
  return t2
end

